""" Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
from time import sleep
from subprocess import Popen, PIPE
import threading
import sys
import schedule
import re
import shlex

from sqlalchemy.exc import DBAPIError

sys.path.append("/opt/cyops-auth")
from handlerworkers import cluster
from utilities.csconstants import CSConstants
from log.cslog import get_ha_logger
from utilities.ha.services import get_cyops_services_list, execute_cmd
from utilities.ha import handler
from utilities.ha.consts import COMMAND_ID_REFRESH_CONFIG
from handlerworkers.apikeys import CSAPIKeys
from utilities.ha.supported_commands import LIST_OF_SUPPORTED_CMDS
import requests
logger = get_ha_logger(__name__)



def build_cluster_command_response(exit_code=0, status_code=200, output="", error=""):
    return {
        "exit_code": exit_code,
        "status_code": status_code,
        "output": output,
        "error": error
    }


def handle_regenerate_appliance_keys_command(uuid):
    CSAPIKeys().save_api_keys_to_files(uuid)
    return 0, '', ''

def check_if_supported_cmd(command):
    is_valid = False
    command = command.strip()
    for cmd_regex in LIST_OF_SUPPORTED_CMDS:
        if re.search(cmd_regex, command):
            is_valid = True
            break
    return is_valid


def api_call_handler(api_call_details):
    method = api_call_details['method']
    headers = api_call_details['headers']
    headers["is-cluster-call"] = "True"
    body = api_call_details['body']
    url = api_call_details['url']

    try:
        if method == 'DELETE' or method == 'GET':
            # When content-lenght header is present, the delete request fails
            headers.pop('content-length', None)
            response_obj = requests.request(method, url, headers=headers)
        else:
            response_obj = requests.request(method, url, data=body, headers=headers)
        if response_obj.status_code != 200:
            return build_cluster_command_response(exit_code=1, status_code=response_obj.status_code)
        else:
            output=""
            if response_obj.text:
                output = response_obj.json()
            return build_cluster_command_response(exit_code=0, status_code=response_obj.status_code, output=output)
    except Exception as error:
        logger.exception(error)
        logger.error("API failed with error: "+str(error))
        return build_cluster_command_response(exit_code=1, status_code=500, error=str(error))


def run_cluster_command_new(pending_command):
    try:
        if pending_command.handler_type == "api-call-handler":
            command_response = api_call_handler(pending_command.payload)
            if command_response["exit_code"] == 0:
                if pending_command.status == CSConstants.CS_COMMAND_TODO_SYNC:
                    cluster.mark_command_complete_new(pending_command.id, command_response)
                else:
                    cluster.dismiss_command(pending_command.id)
            else:
                logger.warn('Command failed with error: ' + str(command_response))
                cluster.mark_command_failed_new(pending_command.id, command_response)
    except Exception as e:
        logger.exception('Error executing command: %s' % pending_command)
        command_response = build_cluster_command_response(exit_code=1, status_code=500, error=str(e))
        cluster.mark_command_failed_new(pending_command.id, command_response)


def run_cluster_command(pending_command):
    commandId = pending_command.id
    command = pending_command.command
    status = pending_command.status
    try:
        output = None
        error = None
        returncode = None
        cluster.mark_command_in_progress(commandId)

        if pending_command.handler_type is not None:
            run_cluster_command_new(pending_command)
            return

        # Keeping below code for backward compatibility

        if command:
            if command.startswith('regenerate-appliance-keys'):
                returncode, output, error = handle_regenerate_appliance_keys_command(command.split(' ')[1])
            else:
                if not check_if_supported_cmd(command):
                    error = "Command [" + command + "] is not whitelisted in framework"
                    logger.warn('Command failed with error: ' + error)
                    cluster.mark_command_failed(commandId, error)
                    return
                command = shlex.split(command)
                logger.info('Executing command ' + str(command))

                command_result = Popen(command, stdout=PIPE, stderr=PIPE)
                output, error = command_result.communicate()
                returncode = command_result.returncode

            if returncode == 0:
                if status == CSConstants.CS_COMMAND_TODO_SYNC:
                    cluster.mark_command_complete(commandId, output.decode('utf-8'))
                else:
                    cluster.dismiss_command(commandId)
            else:
                logger.warn('Command failed with error: '+error.decode('utf-8'))
                cluster.mark_command_failed(commandId, error.decode('utf-8'))
    except Exception as e:
        logger.exception('Error executing command: %s' % command)
        cluster.mark_command_failed(commandId, str(e))


if __name__ == '__main__':
    # ensure that the node is not faulted
    current_node = cluster.get_current_node()
    while not current_node:
        # Case when the node has not yet joined self cluster
        logger.warn("Node is not part of any cluster")
        sleep(20)
        current_node = cluster.get_current_node()
    status = current_node['status']
    if status.lower() == 'faulted':
        logger.error('Node status is faulted. Wait for 2 mins for boot and disable all services.')
        sleep(120)
        serv_list = get_cyops_services_list()
        for serv in serv_list:
            # do not kill self
            if serv.strip() != 'cyops-ha':
                cmd = "systemctl stop " + serv
                execute_cmd(cmd)
                cmd = "systemctl disable " + serv
                execute_cmd(cmd)
                cmd = "systemctl mask " + serv
                execute_cmd(cmd)
        # only ha service keeps running forever
        while True:
            sleep(120)
    else:
        handler.schedule_health_check()
        # create a subprocess for publish_messages listener
        Popen(['/opt/cyops-auth/utilities/ha/cluster_publish.py'])
        while True:
            schedule.run_pending()
            try:
                pending_commands = cluster.get_self_cluster_command()
                for pending_command in pending_commands:
                    if pending_command.commandId == COMMAND_ID_REFRESH_CONFIG:
                        handler.refresh_health_check_schedule()
                        cluster.dismiss_command(pending_command.id)
                        continue
                    # run the command in a separate thread
                    t = threading.Thread(target=run_cluster_command, args=[pending_command])
                    t.daemon = True
                    # TODO: validate that the thread exits once the command completes
                    t.start()
            except DBAPIError as db_error:
                logger.warn("HA service failed to connect to Database")
            except Exception as e:
                logger.exception("Error fetching cluster command")
            # sleep for 10 seconds
            sleep(10)
